# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
import math
import numpy as np
from enum import Enum
from numpy.typing import ArrayLike
from pydantic import BaseModel, ConfigDict
from typing import Optional, Tuple

from ..twobody.utils import findc2c3
from ...constants import MU, SMALL, TWOPI
from ...mathtime.utils import safe_sqrt
from ...mathtime.vector import unit


logger = logging.getLogger(__name__)


# Constants
OOMU = 1 / np.sqrt(MU)


class DirectionOfMotion(Enum):
    LONG = "L"  # Long way
    SHORT = "S"  # Short way


class DirectionOfEnergy(Enum):
    LOW = "L"  # Low
    HIGH = "H"  # High


class DirectionOfFlight(Enum):
    DIRECT = "D"  # Direct motion
    RETROGRADE = "R"  # Retrograde motion


class LambertParams(BaseModel):
    # TODO: use this model directly?
    r1: ArrayLike
    r2: ArrayLike
    v1: Optional[ArrayLike] = None
    dm: Optional[DirectionOfMotion] = None
    de: Optional[DirectionOfEnergy] = None
    df: Optional[DirectionOfFlight] = None
    nrev: Optional[int] = None
    dtsec: Optional[float] = None

    model_config = ConfigDict(arbitrary_types_allowed=True, use_enum_values=True)


########################################################################################
# Supporting Functions
########################################################################################


def calculate_mag_and_angle(r1: ArrayLike, r2: ArrayLike) -> Tuple[float, float, float]:
    """Calculate the magnitudes of two position vectors and the cosine of the
    angle between them.

    Args:
        r1 (array_like): Initial position vector
        r2 (array_like): Final position vector

    Returns:
        tuple: (magr1, magr2, cosdeltanu)
            magr1 (float): Magnitude of the initial position vector
            magr2 (float): Magnitude of the final position vector
            cosdeltanu (float): Cosine of the angle between the two position
    """
    magr1 = float(np.linalg.norm(r1))
    magr2 = float(np.linalg.norm(r2))
    cosdeltanu = np.dot(r1, r2) / (magr1 * magr2)
    cosdeltanu = np.clip(cosdeltanu, -1, 1)  # ensure within bounds [-1, 1]
    return magr1, magr2, float(cosdeltanu)


########################################################################################
# Lambert's Problem
########################################################################################


def min_energy(
    r1: ArrayLike, r2: ArrayLike, dm: DirectionOfMotion, nrev: int
) -> Tuple[np.ndarray, float, float, float]:
    """Solves the Lambert minimum energy problem.

    References:
        Vallado: 2022, p. 479-481, Algorithm 56

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT)
        nrev (int): Number of revolutions (0, 1, 2, ...)

    Returns:
        tuple: (v, aminenergy, tminenergy, tminabs)
            v (np.ndarray): Minimum energy velocity vector in km/s
            aminenergy (float): Minimum energy semi-major axis in km
            tminenergy (float): Minimum energy time of flight in seconds
            tminabs (float): Minimum time of flight (parabolic) in seconds
    """
    # Validate the Pydantic model
    _ = LambertParams(r1=r1, r2=r2, dm=dm, nrev=nrev)

    # Calculate r1 and r2 mag and the cosine of the angle between them
    magr1, magr2, cosdeltanu = calculate_mag_and_angle(r1, r2)

    # Compute the minimum energy semi-major axis
    c = np.sqrt(magr1**2 + magr2**2 - 2 * magr1 * magr2 * cosdeltanu)
    s = 0.5 * (magr1 + magr2 + c)
    aminenergy = 0.5 * s

    # Define alphae and betae
    alphae = np.pi
    betae = 2 * np.arcsin(np.sqrt((s - c) / s))

    # Compute the minimum energy time of flight
    # Use multiplier based on direction of motion
    sign = 1 if dm == DirectionOfMotion.SHORT else -1
    tminenergy = np.sqrt(aminenergy**3 / MU) * (
        2 * nrev * np.pi + alphae + sign * (betae - np.sin(betae))
    )

    # Calculate the parabolic time of flight, which serves as the minimum limit
    tminabs = (1 / 3) * np.sqrt(2 / MU) * (s**1.5 - (s - c) ** 1.5)

    # Compute intermediate values
    rcrossr = np.cross(r1, r2)
    magrcrossr = np.linalg.norm(rcrossr)
    pmin = magr1 * magr2 / c * (1 - cosdeltanu)
    sindeltanu = magrcrossr / (magr1 * magr2) * sign

    # Compute the minimum energy velocity vector
    v = (np.sqrt(MU * pmin) / (magr1 * magr2 * sindeltanu)) * (
        r2 - (1 - magr2 / pmin * (1 - cosdeltanu)) * r1
    )

    return v, aminenergy, tminenergy, tminabs


def min_time(
    r1: ArrayLike,
    r2: ArrayLike,
    dm: DirectionOfMotion,
    de: DirectionOfEnergy,
    nrev: int,
    fa_tol: float = 1e-5,
    fa_iter: int = 20,
) -> Tuple[float, float, float]:
    """Solves Lambert's problem to find the minimum time of flight for the
    multi-revolution cases.

    References:
        Vallado: 2022, p. 481-482, Algorithm 57
        Prussing: JAS 2000

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT)
        de (DirectionOfEnergy): Direction of energy (LOW or HIGH)
        nrev (int): Number of revolutions (0, 1, 2, ...)
        fa_tol (float, optional): Tolerance for the Prussing method min TOF
                                  (defaults to 1e-5)
        fa_iter (int, optional): Maximum number of iterations for the Prussing method
                                 min TOF (defaults to 20)

    Returns:
        tuple: (tmin, tminp, tminenergy)
            tmin (float): Minimum time of flight in seconds
            tminp (float): Minimum time of flight (parabolic) in seconds
            tminenergy (float): Minimum energy time of flight in seconds
    """
    # Validate the Pydantic model
    _ = LambertParams(r1=r1, r2=r2, dm=dm, de=de, nrev=nrev)

    # Create numpy arrays and compute magnitudes of r1 and r2
    magr1, magr2, cosdeltanu = calculate_mag_and_angle(r1, r2)

    # Calculate chord and semiperimeter
    chord = np.sqrt(magr1**2 + magr2**2 - 2 * magr1 * magr2 * cosdeltanu)
    s = (magr1 + magr2 + chord) * 0.5

    # Multipliers based on direction of motion and energy
    sign_dm = 1 if dm == DirectionOfMotion.LONG else -1
    sign_de = 1 if de == DirectionOfEnergy.LOW else -1

    # Calculate minimum parabolic time of flight tasee if orbit is possible
    tminp = (1 / 3) * np.sqrt(2 / MU) * ((s**1.5) + sign_dm * (s - chord) ** 1.5)

    # Calculate minimum energy ellipse time of flight
    amin = 0.5 * s
    beta = 2 * np.arcsin(np.sqrt((s - chord) / s))
    tminenergy = (
        (amin**1.5)
        * ((2 * nrev + 1) * np.pi + sign_dm * (beta - np.sin(beta)))
        / np.sqrt(MU)
    )

    # Iteratively calculate the minimum time of flight (ellipse)
    # using Prussing method (Prussing 1992 AAS, 2000 JAS, Stern 1964 p. 230)
    an = 1.001 * amin
    i = 1
    fa, xi, eta = 10, 0, 0
    while abs(fa) > fa_tol and i <= fa_iter:
        a = an
        alp = 1 / a
        alpha = 2 * np.arcsin(np.sqrt(0.5 * s * alp))
        beta = sign_de * 2 * np.arcsin(np.sqrt(0.5 * (s - chord) * alp))
        xi = alpha - beta
        eta = np.sin(alpha) - np.sin(beta)
        fa = (6 * nrev * np.pi + 3 * xi - eta) * (np.sin(xi) + eta) - 8 * (
            1 - np.cos(xi)
        )
        fadot = (
            (6 * nrev * np.pi + 3 * xi - eta) * (np.cos(xi) + np.cos(alpha))
            + (3 - np.cos(alpha)) * (np.sin(xi) + eta)
            - 8 * np.sin(xi)
        ) * (-alp * np.tan(0.5 * alpha)) + (
            (6 * nrev * np.pi + 3 * xi - eta) * (-np.cos(xi) - np.cos(alpha))
            + (-3 - np.cos(beta)) * (np.sin(xi) + eta)
            + 8 * np.sin(xi)
        ) * (
            -alp * np.tan(0.5 * beta)
        )
        an = a - fa / fadot
        i += 1

    # Calculate the minimum time of flight
    tmin = (an**1.5) * (TWOPI * nrev + xi + sign_dm * eta) / np.sqrt(MU)

    return tmin, tminp, tminenergy


def tmax_rp(
    r1: ArrayLike, r2: ArrayLike, dm: DirectionOfMotion, nrev: int, tol: float = 1e-3
) -> Tuple[float, np.ndarray]:
    """Solves Lambert's problem and finds the TOF for maximum perigee radius.

    References:
        Thompson: 2019

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT)
        nrev (int): Number of revolutions (0, 1, 2, ...)
        tol (float, optional): Distance tolerance in km (defaults to 1e-3, or 1m)

    Returns:
        tuple: (tmaxrp, v1t)
            tmaxrp (float): Maximum perigee time of flight in seconds
            v1t (np.ndarray): Initial velocity vector at r1 in km/s
    """
    v1t = np.zeros(3)
    magr1, magr2 = np.linalg.norm(r1), np.linalg.norm(r2)

    # Calculate the cosine of the angle between the two position vectors
    cos_deltanu = np.dot(r1, r2) / (magr1 * magr2)
    cos_deltanu = np.clip(cos_deltanu, -1, 1)  # ensure within [-1, 1]

    # Calculate the sine of the angle
    rcrossr = np.cross(r1, r2)
    sindeltanu = np.linalg.norm(rcrossr) / (magr1 * magr2)
    if dm == DirectionOfMotion.LONG:
        sindeltanu *= -1  # flip sign for LONG way

    # Calculate the time of flight for maximum perigee radius
    y1, y2 = MU / magr1, MU / magr2
    tempvec = r1 + r2
    if np.linalg.norm(tempvec) < tol:  # nearly circular endpoints
        c = np.sqrt(y1)
        tmaxrp = (TWOPI * nrev + np.arctan2(sindeltanu, cos_deltanu)) * (MU / c**3)
    else:
        if magr1 < magr2:
            c = np.sqrt((y2 - y1 * cos_deltanu) / (1 - cos_deltanu))
            x1, x2 = 0, (y1 - c**2) * sindeltanu
        else:
            c = np.sqrt((y1 - y2 * cos_deltanu) / (1 - cos_deltanu))
            x1, x2 = (-y2 + c**2) * sindeltanu, 0

        r = np.sqrt(x1**2 + (y1 - c**2) ** 2) / c

        # Check if acos is larger than 1
        temp = c * (r**2 + y1 - c**2) / (r * y1)
        e1 = np.arccos(np.clip(temp, -1, 1))
        if x1 < 0:
            e1 = TWOPI - e1

        temp = c * (r**2 + y2 - c**2) / (r * y2)
        e2 = np.arccos(np.clip(temp, -1, 1))
        if x2 < 0:
            e2 = TWOPI - e2
        if e2 < e1:
            e2 += TWOPI

        k = (e2 - e1) - np.sin(e2 - e1)

        tmaxrp = MU * (
            (TWOPI * nrev + k) / np.abs(c**2 - r**2) ** 1.5
            + (c * sindeltanu) / (y1 * y2)
        )

        # Close to 180 deg transfer case
        if magr2 * sindeltanu > tol:
            nunit = unit(rcrossr)
            if sindeltanu < 0:
                nunit *= -1

            nr1 = np.cross(nunit, r1)
            v1t = (x1 / c) * r1 / magr1 + (y1 / c) * (nr1 / magr1)

    return tmaxrp, v1t


########################################################################################
# Battin's Method
########################################################################################


def seebatt(v: float) -> float:
    """Recursively calculates a value used in the Lambert Battin problem using
    pre-defined coefficients.

    Args:
        v (float): Input value for the recursive calculations (v > -1)

    Returns:
        float: The computed value

    Raises:
        ValueError: If `v` is less than -1
    """
    # Check that v is greater than -1
    if v <= -1:
        raise ValueError("Input value v must be greater than -1.")

    # Coefficients derived from Battin's recursive series
    c = [
        9 / 7,
        16 / 63,
        25 / 99,
        36 / 143,
        49 / 195,
        64 / 255,
        81 / 323,
        100 / 399,
        121 / 483,
        144 / 575,
        169 / 675,
        196 / 783,
        225 / 899,
        256 / 1023,
        289 / 1155,
        324 / 1295,
        361 / 1443,
        400 / 1599,
        441 / 1763,
        484 / 1935,
    ]

    # Recursive formulaiton for the Lambert problem
    sqrtopv = np.sqrt(1 + v)
    eta = v / (1 + sqrtopv) ** 2
    ktr = 20
    term2 = 1 + c[ktr - 1] * eta
    for j in range(ktr - 2, -1, -1):
        term2 = 1 + (c[j] * eta / term2)

    return 8 * (1 + sqrtopv) / (3 + (1 / (5 + eta + ((9 / 7) * eta / term2))))


def kbatt(v: float) -> float:
    """Recursively calculates a value used in the Lambert Battin problem using
    pre-defined coefficients.

    Args:
        v (float): Input value for the recursive calculations

    Returns:
        float: The computed value
    """
    # Coefficients derived from Battin's recursive series
    d = [
        1 / 3,
        4 / 27,
        8 / 27,
        2 / 9,
        22 / 81,
        208 / 891,
        340 / 1287,
        418 / 1755,
        598 / 2295,
        700 / 2907,
        928 / 3591,
        1054 / 4347,
        1330 / 5175,
        1480 / 6075,
        1804 / 7047,
        1978 / 8091,
        2350 / 9207,
        2548 / 10395,
        2968 / 11655,
        3190 / 12987,
        3658 / 14391,
    ]

    # Initial values
    sum1, delold, termold = d[0], 1, d[0]

    # Process forwards
    i, ktr = 1, 21
    while i < ktr and abs(termold) > 1e-8:
        del_ = 1 / (1 + d[i] * v * delold)
        term = termold * (del_ - 1)
        sum1 += term
        i += 1
        delold = del_
        termold = term

    # Process backwards
    term2 = 1 + d[-1] * v
    for i in range(ktr - 2):
        sum2 = d[ktr - i - 2] * v / term2
        term2 = 1 + sum2

    return d[0] / term2


def hodograph(
    r1: ArrayLike,
    r2: ArrayLike,
    v1: ArrayLike,
    p: float,
    ecc: float,
    dnu: float,
    dtsec: float,
) -> Tuple[np.ndarray, np.ndarray]:
    """Accomplishes a 180-degree (and 360-degree) transfer for the Lambert problem.

    References:
        Thompson JGCD 2013 v34 n6 1925
        Thompson AAS GNC 2018

    Args:
        r1 (array_like): Initial position vector in km
        r2 (array_like): Final position vector in km
        v1 (array_like): Initial velocity vector in km/s
        p (float): Semi-parameter of transfer orbit in km
        ecc (float): Eccentricity of the transfer orbit
        dnu (float): Change in true anomaly in radians
        dtsec (float): Time between r1 and r2 in seconds

    Returns:
        tuple: (v1t, v2t)
            v1t (np.ndarray): Transfer velocity vector at r1 in km/s
            v2t (np.ndarray): Transfer velocity vector at r2 in km/s
    """
    # Create numpy arrays and compute magnitudes of r1 and r2
    r1, r2 = np.array(r1), np.array(r2)
    magr1, magr2 = np.linalg.norm(r1), np.linalg.norm(r2)

    # Compute parameters a and b
    a = MU * (1 / magr1 - 1 / p)
    b = (MU * ecc / p) ** 2 - a**2

    # Calculate x1 based on b
    x1 = 0 if b <= 0 else -np.sqrt(b)

    # 180-degree or multiple 180-degree transfers
    if abs(np.sin(dnu)) < SMALL:
        # Check that the cross product norm is not zero
        cross_product_r1_v1 = np.cross(r1, v1)
        norm_cross_r1_v1 = np.linalg.norm(cross_product_r1_v1)
        if norm_cross_r1_v1 < SMALL:
            raise ValueError(
                "Vectors r1 and v1 are parallel or nearly parallel;"
                " the vector normal is undefined."
            )

        # Normal vector
        nvec = cross_product_r1_v1 / norm_cross_r1_v1

        # Adjust the direction of x1 based on the time of flight
        if ecc < 1:
            ptx = TWOPI * np.sqrt(p**3 / (MU * (1 - ecc**2) ** 3))
            if dtsec % ptx > ptx * 0.5:
                x1 = -x1
    else:
        # Common path
        y2a = MU / p - x1 * np.sin(dnu) + a * np.cos(dnu)
        y2b = MU / p + x1 * np.sin(dnu) + a * np.cos(dnu)
        if abs(MU / magr2 - y2b) < abs(MU / magr2 - y2a):
            x1 = -x1

        # Check that the cross product norm is not zero
        cross_product_r1_r2 = np.cross(r1, r2)
        norm_cross_r1_r2 = np.linalg.norm(cross_product_r1_r2)
        if norm_cross_r1_r2 < SMALL:
            raise ValueError(
                "Vectors r1 and r2 are parallel or nearly parallel;"
                " the vector normal is undefined."
            )

        # Normal vector
        # Depending on the cross product, this will be normal, in plane, or even a fan
        nvec = cross_product_r1_r2 / norm_cross_r1_r2
        if dnu % TWOPI > np.pi:
            nvec = -nvec

    # Compute transfer velocity vectors
    v1t = (np.sqrt(MU * p) / magr1) * ((x1 / MU) * r1 + np.cross(nvec, r1) / magr1)
    x2 = x1 * np.cos(dnu) + a * np.sin(dnu)
    v2t = (np.sqrt(MU * p) / magr2) * ((x2 / MU) * r2 + np.cross(nvec, r2) / magr2)

    return v1t, v2t


def battin(
    r1: ArrayLike,
    r2: ArrayLike,
    v1: ArrayLike,
    dm: DirectionOfMotion,
    de: DirectionOfEnergy,
    nrev: int,
    dtsec: float,
    n_loops_he: int = 20,
    n_loops_le: int = 30,
) -> Tuple[np.ndarray, np.ndarray]:
    """Solves Lambert's problem using Battin's method.

    This method is developed in battin (1987) and explained by Thompson 2018. It uses
    continued fractions to speed the solution and has several parameters that are
    defined differently than the traditional Gaussian technique.

    References:
        Vallado: 2022, p. 505-510, Algorithm 61
        Battin: 1987, p. 325-342
        Thompson: AAS GNC 2018

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        v1 (array_like): Initial ECI velocity vector in km/s
                         (needed for 180-degree transfer)
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT)
        de (DirectionOfEnergy): Direction of energy (LOW or HIGH)
        nrev (int): Number of revolutions (0, 1, 2, ...)
        dtsec (float): Time between r1 and r2 in seconds
        n_loops_he (int, optional): Number of loops for high energy case
                                    (defaults to 20)
        n_loops_le (int, optional): Number of loops for low energy case
                                    (defaults to 30)

    Returns:
        tuple: (v1t, v2t)
            v1t (np.ndarray): Transfer velocity vector at r1 in km/s
            v2t (np.ndarray): Transfer velocity vector at r2 in km/s

    Notes:
        - The performance over time varies depending on the input TOF (dtsec)
          for the transfer, and Some values of TOF simply have no solutions.
          See plot of time vs. psi (Figure 7-16) in Vallado for more details.
    """
    # Validate the Pydantic model
    _ = LambertParams(r1=r1, v1=v1, r2=r2, dm=dm, de=de, nrev=nrev, dtsec=dtsec)

    # Initialize values
    v1t, v2t = np.array([np.NAN] * 3), np.array([np.NAN] * 3)
    y = 0

    # Create numpy arrays and compute magnitudes of r1 and r2
    magr1, magr2, cosdeltanu = calculate_mag_and_angle(r1, r2)

    # Determine direction of flight
    magrcrossr = np.linalg.norm(np.cross(r1, r2))
    sign = 1 if dm == DirectionOfMotion.SHORT else -1
    sindeltanu = sign * magrcrossr / (magr1 * magr2)

    # Compute delta nu
    dnu = np.arctan2(sindeltanu, cosdeltanu)
    dnu = dnu if dnu >= 0 else TWOPI + dnu  # Ensure positive angle

    # Calculate chord and semiperimeter
    chord = np.sqrt(magr1**2 + magr2**2 - 2 * magr1 * magr2 * cosdeltanu)
    s = (magr1 + magr2 + chord) * 0.5
    ror = magr2 / magr1
    eps = ror - 1

    # Calculate lambda, L, and m
    lam = 1 / s * np.sqrt(magr1 * magr2) * np.cos(dnu * 0.5)
    l_ = ((1 - lam) / (1 + lam)) ** 2
    m = 8 * MU * dtsec**2 / (s**3 * (1 + lam) ** 6)

    # Initial guess for x
    xn = 1 + 4 * l_ if nrev > 0 else l_

    # Context for the safe square root function errors
    con = "Battin's method intermediate calculations: please adjust `dtsec`"

    # Default values for y and xn
    y_default, xn_default = 75, 1
    warn_msg = f"Failed to calculate y, setting to {y_default} and xn to {xn_default}"

    # High energy case adjustments for long way, retrograde multi-rev
    if de == DirectionOfEnergy.HIGH and nrev > 0:
        xn, x = 1e-20, 10
        loops = 1
        while abs(xn - x) >= SMALL and loops <= n_loops_he:
            # Calculate h1 and h2
            x = xn
            temp = 1 / (2 * (l_ - x**2))
            temp1 = safe_sqrt(x, con)
            temp2 = (nrev * np.pi * 0.5 + np.arctan(temp1)) / temp1
            h1 = temp * (l_ + x) * (1 + 2 * x + l_)
            h2 = temp * m * temp1 * ((l_ - x**2) * temp2 - (l_ + x))

            # Calculate b and f
            b = 0.25 * 27 * h2 / ((temp1 * (1 + h1)) ** 3)
            if b < 0:
                f = 2 * np.cos(1 / 3 * np.arccos(safe_sqrt(b + 1), con))
            else:
                a_ = (safe_sqrt(b, con) + safe_sqrt(b + 1, con)) ** (1 / 3)
                f = a_ + 1 / a_

            # Calculate y and xn
            y = 2 / 3 * temp1 * (1 + h1) * (safe_sqrt(b + 1, con) / f + 1)
            xn = 0.5 * (
                (m / (y**2) - (1 + l_))
                - safe_sqrt((m / (y**2) - (1 + l_)) ** 2 - 4 * l_, con)
            )

            # Check for NaN values
            if np.isnan(y):
                y, xn = y_default, xn_default
                logger.warning(warn_msg)

            loops += 1

        # Determine transfer velocity vectors for high energy case
        x = xn
        a = s * (1 + lam) ** 2 * (1 + x) * (l_ + x) / (8 * x)
        p = (2 * magr1 * magr2 * (1 + x) * np.sin(dnu * 0.5) ** 2) / (
            s * (1 + lam) ** 2 * (l_ + x)
        )
        ecc = safe_sqrt(1 - p / a, con)
        v1t, v2t = hodograph(r1, r2, v1, p, ecc, dnu, dtsec)
    else:
        # Standard processing, low energy case
        loops, x = 1, 10
        while abs(xn - x) >= SMALL and loops <= n_loops_le:
            # Calculate h1 and h2
            x = xn
            if nrev > 0:
                temp = 1 / ((1 + 2 * x + l_) * (4 * x**2))
                temp1 = (nrev * np.pi * 0.5 + np.arctan(safe_sqrt(x, con))) / safe_sqrt(
                    x, con
                )
                h1 = temp * (l_ + x) ** 2 * (3 * (1 + x) ** 2 * temp1 - (3 + 5 * x))
                h2 = temp * m * ((x**2 - x * (1 + l_) - 3 * l_) * temp1 + (3 * l_ + x))
            else:
                tempx = seebatt(x)
                denom = 1 / ((1 + 2 * x + l_) * (4 * x + tempx * (3 + x)))
                h1 = (l_ + x) ** 2 * (1 + 3 * x + tempx) * denom
                h2 = m * (x - l_ + tempx) * denom

            # Calculate y and xn
            b = 0.25 * 27 * h2 / ((1 + h1) ** 3)
            u = 0.5 * b / (1 + safe_sqrt(1 + b, con))
            k2 = kbatt(u)
            y = ((1 + h1) / 3) * (2 + safe_sqrt(1 + b, con) / (1 + 2 * u * k2 * k2))
            xn = safe_sqrt(((1 - l_) * 0.5) ** 2 + m / (y**2), con) - (1 + l_) * 0.5

            # Check for NaN values
            if np.isnan(y):
                y, xn = y_default, xn_default
                logger.warning(warn_msg)

            loops += 1

        # Determine transfer velocity vectors for standard case
        if loops < n_loops_le:
            p = (2 * magr1 * magr2 * y**2 * (1 + x) ** 2 * np.sin(dnu * 0.5) ** 2) / (
                m * s * (1 + lam) ** 2
            )
            ecc = safe_sqrt(
                (
                    eps**2
                    + 4
                    * magr2
                    / magr1
                    * np.sin(dnu * 0.5) ** 2
                    * ((l_ - x) / (l_ + x)) ** 2
                )
                / (eps**2 + 4 * magr2 / magr1 * np.sin(dnu * 0.5) ** 2),
                con,
            )
            v1t, v2t = hodograph(r1, r2, v1, p, ecc, dnu, dtsec)

    return v1t, v2t


########################################################################################
# Universal Variable Lambert Problem
########################################################################################


def _calculate_c2dot_c3dot(
    psi: float, c2: float, c3: float, tol: float = 1e-5
) -> Tuple[float, float, float, float]:
    """Calculate the derivatives of c2 and c3 with respect to psi.

    Args:
        psi (float): Psi value
        c2 (float): Coefficient c2
        c3 (float): Coefficient c3
        tol (float, optional): Tolerance for small psi (defaults to 1e-5)

    Returns:
        tuple: (c2dot, c3dot, c2ddot, c3ddot)
            c2dot (float): Derivative of c2 with respect to psi
            c3dot (float): Derivative of c3 with respect to psi
            c2ddot (float): Second derivative of c2 with respect to psi
            c3ddot (float): Second derivative of c3 with respect to psi
    """
    if abs(psi) > tol:
        c2dot = 0.5 / psi * (1 - psi * c3 - 2 * c2)
        c3dot = 0.5 / psi * (c2 - 3 * c3)
        c2ddot = 1 / (4 * psi**2) * ((8 - psi) * c2 + 5 * psi * c3 - 4)
        c3ddot = 1 / (4 * psi**2) * ((15 - psi) * c3 - 7 * c2 + 1)
    else:
        # Taylor series expansion for small psi
        c2dot = sum(
            (-1) ** (i + 1) * (i + 1) * psi**i / math.factorial(2 * i + 4)
            for i in range(5)
        )
        c3dot = sum(
            (-1) ** (i + 1) * (i + 1) * psi**i / math.factorial(2 * i + 5)
            for i in range(5)
        )
        c2ddot, c3ddot = 0, 0

    return c2dot, c3dot, c2ddot, c3ddot


def _calculate_dtdpsi(
    x: float, c2: float, c3: float, c2dot: float, c3dot: float, vara: float, y: float
) -> float:
    """Calculate the derivative of time of flight with respect to psi.

    Args:
        x (float): Value of x
        c2 (float): Coefficient c2
        c3 (float): Coefficient c3
        c2dot (float): Derivative of c2 with respect to psi
        c3dot (float): Derivative of c3 with respect to psi
        vara (float): Value of vara
        y (float): Value of y

    Returns:
        float: Derivative of time of flight with respect to psi
    """
    return (
        x**3 * (c3dot - 3 * c3 * c2dot / (2 * c2))
        + 0.125 * vara * (3 * c3 * np.sqrt(y) / c2 + vara / x)
    ) * OOMU


def get_kbiu(r1: ArrayLike, r2: ArrayLike, order: int) -> Tuple[np.ndarray, np.ndarray]:
    """Form the minimum time and universal variable matrix for multi-rev cases.

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        order (int): The number of revolutions to consider

    Returns:
        tuple: (kbi_arr, kbil_arr)
            kbi_arr (np.ndarray): Matrix of kbi and time of flight for SHORT direction
            kbil_arr (np.ndarray): Matrix of kbi and time of flight for LONG direction
    """
    kbi_arr = np.zeros((order, 2))
    for i in range(order):
        kbi, tof = universal_min(r1, r2, DirectionOfMotion.SHORT, i + 1)
        kbi_arr[i, 0] = kbi
        kbi_arr[i, 1] = tof

    kbil_arr = np.zeros((order, 2))
    for i in range(order):
        kbil, tofl = universal_min(r1, r2, DirectionOfMotion.LONG, i + 1)
        kbil_arr[i, 0] = kbil
        kbil_arr[i, 1] = tofl

    return kbi_arr, kbil_arr


def universal_min(
    r1: ArrayLike, r2: ArrayLike, dm: DirectionOfMotion, nrev: int, n_iter: int = 20
) -> Tuple[float, float]:
    """Find the minimum kbi value for the universal variable Lambert problem for the
    multi-rev cases.

    References:
        Arora and Russell: AAS 10-198

    Args:
        r1 (array_like): Initial ECI position vector in km
        r2 (array_like): Final ECI position vector in km
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT)
        nrev (int): Number of revolutions (0, 1, 2, ...)
        n_iter (int): Number of iterations to perform (defaults to 20)

    Returns:
        tuple: (kbi, tof)
            kbi (float): K value for min TOF
            tof (float): Time of flight in seconds

    TODO:
        - Identify and capture any exceptions that may occur due to bad inputs
    """
    # Validate the Pydantic model
    _ = LambertParams(r1=r1, r2=r2, dm=dm, nrev=nrev)

    # Create numpy arrays and compute magnitudes of r1 and r2
    magr1, magr2, cosdeltanu = calculate_mag_and_angle(r1, r2)

    # Determine vara based on direction of motion
    sign_dm = -1 if dm == DirectionOfMotion.LONG else 1
    vara = sign_dm * np.sqrt(magr1 * magr2 * (1.0 + cosdeltanu))

    # Outer bounds for the nrev case
    lower = 4 * nrev**2 * np.pi**2
    upper = 4 * (nrev + 1) ** 2 * np.pi**2

    # Streamline by narrowing down the bounds (since it's near the center)
    upper = lower + (upper - lower) * 0.6
    lower = lower + (upper - lower) * 0.3

    # Initial psi guess (put in center of bounds)
    psiold = (upper + lower) * 0.5

    # Get initial values of c2 and c3
    c2, c3 = findc2c3(psiold)

    # Iterative loop to find minimum psi
    x, sqrty, loops = 0, 0, 0
    dtdpsi, psinew = 200, psiold
    while abs(dtdpsi) >= 0.1 and loops < n_iter:
        # Calculate y and x
        if abs(c2) > SMALL:
            y = magr1 + magr2 - (vara * (1 - psiold * c3) / np.sqrt(c2))
        else:
            y = magr1 + magr2
        x = np.sqrt(y / c2) if abs(c2) > SMALL else 0

        # Calculate derivatives of c2 and c3
        c2dot, c3dot, c2ddot, c3ddot = _calculate_c2dot_c3dot(psiold, c2, c3)

        # Solve for dt = 0.0
        dtdpsi = _calculate_dtdpsi(x, c2, c3, c2dot, c3dot, vara, y)

        # Solve for second derivative of dt with respect to psi
        sqrty = np.sqrt(y)
        q = 0.25 * vara * np.sqrt(c2) - x**2 * c2dot
        s1 = -24 * q * x**3 * c2 * sqrty * c3dot
        s2 = 36 * q * x**3 * sqrty * c3 * c2dot - 16 * x**5 * sqrty * c3ddot * c2**2
        s3 = (
            24 * x**5 * sqrty * (c3dot * c2dot * c2 + c3 * c2ddot * c2 - c3 * c2dot**2)
            - 6 * vara * c3dot * y * c2 * x**2
        )
        s4 = (
            -0.75 * vara**2 * c3 * c2**1.5 * x**2
            + 6 * vara * c3 * y * c2dot * x**2
            + (vara**2 * c2 * (0.25 * vara * np.sqrt(c2) - x**2 * c2)) * sqrty / x
        )
        dtdpsi2 = -(s1 + s2 + s3 + s4) / (16 * np.sqrt(MU) * (c2**2 * sqrty * x**2))

        # Newton-Raphson update for psi
        psinew = psiold - dtdpsi / dtdpsi2

        # Update psi and c2, c3 for the next iteration
        psiold = psinew
        c2, c3 = findc2c3(psiold)
        loops += 1

    # Calculate time of flight and final psi
    dtnew = (x**3 * c3 + vara * sqrty) * OOMU
    tof = dtnew
    kbi = psinew

    return kbi, tof


def universal(
    r1: ArrayLike,
    v1: ArrayLike,
    r2: ArrayLike,
    dtsec: float,
    dm: DirectionOfMotion,
    de: DirectionOfEnergy,
    nrev: int,
    kbi: float,
    tol: float = 1e-05,
    n_iter: int = 20,
) -> Tuple[np.ndarray, np.ndarray]:
    """Solves Lambert's problem for orbit determination, returns the velocity vectors at
    each of two given position vectors.

    The solution uses universal variables for calculation and a bissection technique
    updating psi.

    References:
        Vallado: 2022, p. 499-505, Algorithm 60

    Args:
        r1 (array_like): Initial position vector in km
        v1 (array_like): Initial velocity vector in km/s
        r2 (array_like): Final position vector in km
        dtsec (float): Time of flight in seconds
        dm (DirectionOfMotion): Direction of motion (LONG or SHORT period)
        de (DirectionOfEnergy): Direction of energy (HIGH or LOW)
        nrev (int): Number of revolutions (0, 1, 2, ...)
        kbi (float): K value for the minimum time of flight
        tol (float): Tolerance for the Lambert problem (defaults to 1e-05)
                     (can affect cases where znew is multiples of 2pi^2)
        n_iter (int): Maximum number of iterations to perform (defaults to 20)

    Returns:
        tuple: (v1dv, v2dv)
            v1dv (np.ndarray): Transfer ECI velocity vector at r1 in km/s
            v2dv (np.ndarray): Transfer ECI velocity vector at r2 in km/s

    Notes:
        - If the orbit is not possible, the method will log an error and return the
          Battin method or a Hohmann transfer
        - This method is sensitive to inputs, specifically `dtsec` and `psi` values in
          `psi_vec` (for multi-rev cases) - a bad combination can lead to no solutions
          (see Vallado 2022, Figure 7-16)
    """
    # Validate the Pydantic model
    _ = LambertParams(r1=r1, v1=v1, r2=r2, dtsec=dtsec, dm=dm, de=de, nrev=nrev)

    # Definitions and initialization
    max_ynegktr_iters = 10  # maximum number of iterations for y < 0
    v1dv, v2dv = np.zeros(3), np.zeros(3)

    # Compute magnitudes of r1 and r2
    magr1, magr2, cosdeltanu = calculate_mag_and_angle(r1, r2)

    # Determine vara based on direction of motion
    sign = -1 if dm == DirectionOfMotion.LONG else 1
    vara = sign * np.sqrt(magr1 * magr2 * (1 + cosdeltanu))

    # Set up initial bounds for bisection
    if nrev == 0:
        # Hyperbolic and parabolic solutions
        lower = -16 * np.pi**2
        upper = 4 * np.pi**2
    else:
        lower = 4 * nrev**2 * np.pi**2
        upper = 4 * (nrev + 1) ** 2 * np.pi**2
        if de == DirectionOfEnergy.HIGH:
            upper = kbi
        else:
            lower = kbi

    # Form initial guess for psi
    if nrev == 0:
        # From empirical data (Arora and Russell, 2010)
        psiold = (np.log(dtsec) - 9.61202327) / 0.10918231
        psiold = min(psiold, upper - np.pi)
    else:
        psiold = lower + (upper - lower) * 0.5

    # Initial values for c2 and c3
    c2new, c3new = findc2c3(psiold)

    # Compute initial y and dtold
    if abs(c2new) > tol:
        y = magr1 + magr2 - (vara * (1 - psiold * c3new) / np.sqrt(c2new))
    else:
        y = magr1 + magr2

    xold = np.sqrt(y / c2new) if abs(c2new) > tol else 0
    dtold = (xold**3 * c3new + vara * np.sqrt(y)) * OOMU

    # Check if orbit is not possible
    if abs(vara) < 0.2:  # not exactly zero
        # Can't do bissection because w series is not accurate
        logger.error("Orbit is not possible")

        # Call Battin method or use a Hohmann transfer in 3D
        atx = (MU * (dtsec / np.pi) ** 2) ** (1 / 3)  # half period
        v1tmag = np.sqrt(2 * MU / magr1 - MU / atx)
        v2tmag = np.sqrt(2 * MU / magr2 - MU / atx)

        # Compute the direction of the velocity vectors
        wxu = unit(np.cross(r1, v1))
        v1diru = unit(np.cross(r1, wxu))
        v2diru = unit(np.cross(r2, wxu))

        # Compute the velocities using the direction and magnitude
        v1dv = -v1tmag * v1diru
        v2dv = -v2tmag * v2diru

        return v1dv, v2dv

    # Loop for iteration
    loops, ynegktr, dtnew = 0, 1, -10

    # Loop to find psi
    while abs(dtnew - dtsec) >= tol and loops < n_iter and ynegktr <= max_ynegktr_iters:
        if abs(c2new) > tol:
            y = magr1 + magr2 - (vara * (1 - psiold * c3new) / np.sqrt(c2new))
        else:
            y = magr1 + magr2

        # Check for negative y values
        if y < 0 < vara:
            while y < 0 and ynegktr < max_ynegktr_iters:
                psinew = (
                    0.8 * (1 / c3new) * (1 - (magr1 + magr2) * np.sqrt(c2new) / vara)
                )
                c2new, c3new = findc2c3(psinew)
                psiold = psinew
                lower = psiold
                y = magr1 + magr2 - (vara * (1 - psiold * c3new) / np.sqrt(c2new))
                ynegktr += 1

        loops += 1

        # Check for convergence
        if ynegktr < max_ynegktr_iters:
            xold = np.sqrt(y / c2new) if abs(c2new) > tol else 0
            dtnew = (xold**3 * c3new + vara * np.sqrt(y)) * OOMU

            # Newton-Raphson iteration for psi update
            c2dot, c3dot, _, _ = _calculate_c2dot_c3dot(psiold, c2new, c3new)

            # Calculate new psi
            dtdpsi = _calculate_dtdpsi(xold, c2new, c3new, c2dot, c3dot, vara, y)
            psinew = psiold - (dtnew - dtsec) / dtdpsi

            # Newton iteration test to see if it keeps within the bounds
            if psinew > upper or psinew < lower:
                # psi is outside bounds (too steep a slope)
                # Re-adjust upper and lower bounds
                if de == DirectionOfEnergy.LOW or nrev == 0:
                    if dtold < dtsec:
                        lower = psiold
                    else:
                        upper = psiold
                else:
                    if dtold < dtsec:
                        upper = psiold
                    else:
                        lower = psiold

                psinew = (upper + lower) * 0.5

            # Update c2 and c3
            c2new, c3new = findc2c3(psinew)
            psiold = psinew
            dtold = dtnew

            # Make sure the first guess isn't too close
            if abs(dtnew - dtsec) < tol and loops == 1:
                dtnew = dtsec - 1

    # Check for non-convergence
    if loops >= n_iter or ynegktr >= max_ynegktr_iters:
        logger.error("Lambert did not converge! Try different values of dtsec or psi")
        if ynegktr >= max_ynegktr_iters:
            logger.error("y is negative")
    else:
        # Compute velocity vectors using f and g series
        f = 1 - y / magr1
        gdot = 1 - y / magr2
        g = 1 / (vara * np.sqrt(y / MU))

        for i in range(3):
            v1dv[i] = (r2[i] - f * r1[i]) * g
            v2dv[i] = (gdot * r2[i] - r1[i]) * g

    return v1dv, v2dv



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------


import numpy as np
from numpy.typing import ArrayLike
from typing import Tuple

from . import iau_transform as iau
from .data import IAU80Array, IAU06Array, IAU06xysArray
from .sidereal import gstime, sidereal
from .utils import precess, nutation, polarm
from ... import constants as const


def calc_omegaearth(lod: float) -> np.ndarray:
    """Calculates the Earth's rotation vector.

    Args:
        lod (float): Excess length of day in seconds

    Returns:
        np.ndarray: Earth's rotation vector in rad/s
    """
    return np.array([0, 0, const.EARTHROT * (1 - lod / const.DAY2SEC)])


def calc_orbit_effects(
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
    use_iau80: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Calculates the orbit effects from precession, nutation, sidereal time, and polar
    motion.

    Args:
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)
        use_iau80 (bool, optional): Use IAU 1980 model for precession/nutation
                                    (default True)

    Returns:
        tuple: (prec, nut, st, pm, omegaearth)
            prec (np.ndarray): Transformation matrix for MOD to J2000
            nut (np.ndarray): Transformation matrix for TOD - MOD
            st (np.ndarray): Transformation matrix for PEF to TOD
            pm (np.ndarray): Transformation matrix for ECEF to PEF
            omegaearth (np.ndarray): Earth angular rotation vecctor
    """
    # Find matrices that account for various orbit effects
    prec, *_ = precess(ttt, opt="80")
    deltapsi, _, meaneps, omega, nut = nutation(ttt, ddpsi, ddeps, iau80arr)
    st, _ = sidereal(jdut1, deltapsi, meaneps, omega, lod, use_iau80, eqeterms)
    pm = polarm(xp, yp, ttt, use_iau80=use_iau80)

    # Calculate the effects of Earth's rotation
    omegaearth = calc_omegaearth(lod)

    return prec, nut, st, pm, omegaearth


########################################################################################
# ECI <-> ECEF Frame Conversions
########################################################################################


def compute_iau06_matrices(
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float,
    ddy: float,
    use_full_series: bool,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Computes the precession/nutation matrix, sidereal time matrix, polar motion
    matrix, and Earth rotation vector.

    Args:
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians
        ddy (float, optional): EOP correction for y in radians
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (pnb, st, pm, omegaearth)
            pnb (np.ndarray): Precession/nutation matrix
            st (np.ndarray): Sidereal time matrix
            pm (np.ndarray): Polar motion matrix
            omegaearth (np.ndarray): Earth rotation vector
    """
    # Calculate precession-nutation and sidereal matrices
    *_, pnb = iau.iau06xys(ttt, iau06arr, ddx, ddy, iau06xysarr, use_full_series)
    st, _ = sidereal(jdut1, 0, 0, 0, lod, use_iau80=False)

    # Earth rotation
    omegaearth = calc_omegaearth(lod)

    # Polar motion matrix
    pm = polarm(xp, yp, ttt, use_iau80=False)

    return pnb, st, pm, omegaearth


def eci2ecef(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox frame (J2000) to the
    Earth-fixed frame (ECEF).

    References:
        Vallado: 2022, p. 223-230

    Args:
        reci (array_like): ECi position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector
            vecef (np.ndarray): ECEF velocity vector
            aecef (np.ndarray): ECEF acceleration vector

    TODO: The acceleration transformation is not correct and needs to be fixed.
    """
    # Find matrices that account for various orbit effects
    prec, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Position transformation
    rpef = st.T @ nut.T @ prec.T @ reci
    recef = pm.T @ rpef

    # Velocity transformation
    vpef = st.T @ nut.T @ prec.T @ veci - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Acceleration transformation
    aecef = (
        pm.T @ (st.T @ nut.T @ prec.T @ aeci)
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


def eci2ecef06(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI frame (J2000) to the Earth-fixed frame (ECEF)
    using the IAU 2006 xys approach.

    References:
        Vallado, 2022, p. 223-230

    Args:
        reci (ArrayLike): ECI position vector in km
        veci (ArrayLike): ECI velocity vector in km/s
        aeci (ArrayLike): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data
                                          (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Compute the IAU06 matrices
    pnb, st, pm, omegaearth = compute_iau06_matrices(
        ttt, jdut1, lod, xp, yp, iau06arr, iau06xysarr, ddx, ddy, use_full_series
    )

    # Transform position
    rpef = st.T @ pnb.T @ reci
    recef = pm.T @ rpef

    # Transform velocity
    vpef = st.T @ pnb.T @ veci - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Transform acceleration
    aecef = pm.T @ (
        st.T @ pnb.T @ aeci
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


def ecef2eci(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed frame (ECEF) to the ECI mean equator,
    mean equinox (J2000) frame.

    References:
        Vallado: 2022, p. 223-230

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    prec, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Position transformation
    rpef = np.dot(pm, recef)
    reci = np.dot(prec, np.dot(nut, np.dot(st, rpef)))

    # Velocity transformation
    vpef = np.dot(pm, vecef)
    veci = np.dot(prec, np.dot(nut, np.dot(st, vpef + np.cross(omegaearth, rpef))))

    # Acceleration transformation
    aeci = (
        np.dot(prec, np.dot(nut, np.dot(st, np.dot(pm, aecef))))
        + np.cross(omegaearth, np.cross(omegaearth, rpef))
        + 2 * np.cross(omegaearth, vpef)
    )

    return reci, veci, aeci


def ecef2eci06(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed frame (ECEF) to the ECI (J2000) frame.

    References:
        Vallado, 2022, p. 223-230

    Args:
        recef (ArrayLike): ECEF position vector in km
        vecef (ArrayLike): ECEF velocity vector in km/s
        aecef (ArrayLike): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Compute the IAU06 matrices
    pnb, st, pm, omegaearth = compute_iau06_matrices(
        ttt, jdut1, lod, xp, yp, iau06arr, iau06xysarr, ddx, ddy, use_full_series
    )

    # Transform position
    rpef = pm @ recef
    reci = pnb @ st @ rpef

    # Transform velocity
    vpef = pm @ vecef
    veci = pnb @ st @ (vpef + np.cross(omegaearth, rpef))

    # Transform acceleration
    temp = np.cross(omegaearth, rpef)
    aeci = (
        pnb
        @ st
        @ (pm @ aecef + np.cross(omegaearth, temp) + 2 * np.cross(omegaearth, vpef))
    )

    return reci, veci, aeci


########################################################################################
# ECI <-> PEF Frame Conversions
########################################################################################


def eci2pef(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox frame (J2000), to the
    pseudo Earth-fixed frame (PEF).

    References:
        Vallado: 2022, p. 224

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECi velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        ddpsi (float): Nutation correction for delta psi in radians
        ddeps (float): Nutation correction for delta epsilon in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool): Add terms for ast calculation (default True)

    Returns:
        tuple: (rpef, vpef, apef)
            rpef (np.ndarray): PEF position vector in km
            vpef (np.ndarray): PEF velocity vector in km/s
            apef (np.ndarray): PEF acceleration vector in km/s²
    """
    # Compute the IAU matrices
    prec, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, 0, 0, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform vectors
    rpef = st.T @ nut.T @ prec.T @ reci
    vpef = st.T @ nut.T @ prec.T @ veci - np.cross(omegaearth, rpef)
    apef = (
        st.T @ nut.T @ prec.T @ aeci
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return rpef, vpef, apef


def pef2eci(
    rpef: ArrayLike,
    vpef: ArrayLike,
    apef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the pseudo Earth-fixed (PEF) frame to the ECI mean
    equator, mean equinox (J2000) frame using the IAU 1980 model.

    References:
        Vallado: 2022, p. 224

    Args:
        rpef (array_like): PEf position vector in km
        vpef (array_like): PEF velocity vector in km/s
        apef (array_like): PEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    prec, nut, st, _, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, 0, 0, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform vectors
    reci = prec @ nut @ st @ np.asarray(rpef)
    veci = prec @ nut @ st @ (np.asarray(vpef) + np.cross(omegaearth, rpef))
    aeci = (
        prec
        @ nut
        @ st
        @ (
            np.asarray(apef)
            + np.cross(omegaearth, np.cross(omegaearth, rpef))
            + 2 * np.cross(omegaearth, vpef)
        )
    )

    return reci, veci, aeci


########################################################################################
# ECI <-> TOD Frame Conversions
########################################################################################


def eci2tod(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox frame (J2000) to the
    true equator, true equinox of date (TOD) frame.

    References:
        Vallado: 2022, p. 225

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation

    Returns:
        tuple: (rtod, vtod, atod)
            rtod (np.ndarray): TOD position vector in km
            vtod (np.ndarray): TOD velocity vector in km/s
            atod (np.ndarray): TOD acceleration vector in km/s²
    """
    # Precession (IAU 1980 model)
    prec, *_ = precess(ttt, opt="80")

    # Nutation
    *_, nut = nutation(ttt, ddpsi, ddeps, iau80arr)

    # Transform vectors
    rtod = nut.T @ prec.T @ np.asarray(reci)
    vtod = nut.T @ prec.T @ np.asarray(veci)
    atod = nut.T @ prec.T @ np.asarray(aeci)

    return rtod, vtod, atod


def tod2eci(
    rtod: ArrayLike,
    vtod: ArrayLike,
    atod: ArrayLike,
    ttt: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the true equator, true equinox of date (TOD) frame to
    the ECI mean equator, mean equinox (J2000) frame.

    References:
        Vallado: 2022, p. 225

    Args:
        rtod (array_like): TOD position vector in km
        vtod (array_like): TOD velocity vector in km/s
        atod (array_like): TOD acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Precession (IAU 1980 model)
    prec, *_ = precess(ttt, opt="80")

    # Nutation
    *_, nut = nutation(ttt, ddpsi, ddeps, iau80arr)

    # Transform vectors
    reci = prec @ nut @ np.asarray(rtod)
    veci = prec @ nut @ np.asarray(vtod)
    aeci = prec @ nut @ np.asarray(atod)

    return reci, veci, aeci


########################################################################################
# ECI <-> MOD Frame Conversions
########################################################################################


def eci2mod(
    reci: ArrayLike, veci: ArrayLike, aeci: ArrayLike, ttt: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox frame (J2000) to the
    mean equator, mean equinox of date (MOD) frame.

    References:
        Vallado: 2022, p. 227

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT

    Returns:
        tuple: (rmod, vmod, amod)
            rmod (np.ndarray): MOD position vector in km
            vmod (np.ndarray): MOD velocity vector in km/s
            amod (np.ndarray): MOD acceleration vector in km/s²
    """
    # Precession (IAU 1980 model)
    prec, *_ = precess(ttt, opt="80")

    # Transform vectors
    rmod = prec.T @ np.asarray(reci)
    vmod = prec.T @ np.asarray(veci)
    amod = prec.T @ np.asarray(aeci)

    return rmod, vmod, amod


def mod2eci(
    rmod: ArrayLike, vmod: ArrayLike, amod: ArrayLike, ttt: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the mean equator, mean equinox of date (MOD) frame
    to the ECI mean equator, mean equinox (J2000) frame.

    References:
        Vallado: 2022, p. 227

    Args:
        rmod (array_like): MOD position vector in km
        vmod (array_like): MOD velocity vector in km/s
        amod (array_like): MOD acceleration vector in km/s²
        ttt (float): Julian centuries of TT

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Precession (IAU 1980 model)
    prec, *_ = precess(ttt, opt="80")

    # Transform vectors
    reci = prec @ np.asarray(rmod)
    veci = prec @ np.asarray(vmod)
    aeci = prec @ np.asarray(amod)

    return reci, veci, aeci


########################################################################################
# ECI <-> TEME Frame Conversions
########################################################################################


def _get_teme_eci_transform(
    ttt: float, ddpsi: float, ddeps: float, iau80arr: IAU80Array
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Generates the transformation matrices for TEME ↔ ECI.

    Args:
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation

    Returns:
        tuple: (prec, nut, eqe)
            prec (np.ndarray): Precession matrix
            nut (np.ndarray): Nutation matrix
            eqe (np.ndarray): Equation of equinoxes rotation matrix
    """
    # Precession (IAU 1980 model)
    prec, *_ = precess(ttt, opt="80")

    # Nutation
    deltapsi, _, meaneps, _, nut = nutation(ttt, ddpsi, ddeps, iau80arr)

    # Equation of equinoxes (geometric terms only)
    eqeg = deltapsi * np.cos(meaneps)
    eqeg = np.remainder(eqeg, const.TWOPI)

    # Construct the rotation matrix for the equation of equinoxes
    eqe = np.array(
        [[np.cos(eqeg), np.sin(eqeg), 0], [-np.sin(eqeg), np.cos(eqeg), 0], [0, 0, 1]]
    )

    return prec, nut, eqe


def eci2teme(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox (J2000) frame to the
    true equator, mean equinox (TEME) frame.

    References:
        Vallado: 2022, p. 232-234

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation

    Returns:
        tuple: (rteme, vteme, ateme)
            rteme (np.ndarray): TEME position vector in km
            vteme (np.ndarray): TEME velocity vector in km/s
            ateme (np.ndarray): TEME acceleration vector in km/s²
    """
    # Get the individual transformation matrices
    prec, nut, eqe = _get_teme_eci_transform(ttt, ddpsi, ddeps, iau80arr)

    # Combined transformation matrix
    tm = eqe @ nut.T @ prec.T

    # Transform vectors
    rteme = tm @ np.asarray(reci)
    vteme = tm @ np.asarray(veci)
    ateme = tm @ np.asarray(aeci)

    return rteme, vteme, ateme


def teme2eci(
    rteme: ArrayLike,
    vteme: ArrayLike,
    ateme: ArrayLike,
    ttt: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the true equator, mean equinox (TEME) frame to the ECI
    mean equator, mean equinox (J2000) frame.

    References:
        Vallado: 2022, p. 232-234

    Args:
        rteme (array_like): TEME position vector in km
        vteme (array_like): TEME velocity vector in km/s
        ateme (array_like): TEME acceleration vector in km/s²
                            (set to zeros if not available)
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Get the individual transformation matrices
    prec, nut, eqe = _get_teme_eci_transform(ttt, ddpsi, ddeps, iau80arr)

    # Combined transformation matrix
    tm = prec @ nut @ eqe.T

    # Transform vectors
    reci = tm @ np.asarray(rteme)
    veci = tm @ np.asarray(vteme)
    aeci = tm @ np.asarray(ateme)

    return reci, veci, aeci


########################################################################################
# ECI <-> CIRS Frame Conversions
########################################################################################


def eci2cirs(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox (J2000) frame to the
    Celestial Intermediate Reference System (CIRS) frame using the XYS approach.

    References:
        Vallado: 2022, p. 214

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (rcirs, vcirs, acirs)
            rcirs (np.ndarray): CIRS position vector in km
            vcirs (np.ndarray): CIRS velocity vector in km/s
            acirs (np.ndarray): CIRS acceleration vector in km/s²
    """
    # Compute transformation matrix using XYS approach
    *_, pnb = iau.iau06xys(ttt, iau06arr, ddx, ddy, iau06xysarr, use_full_series)

    # Transform vectors
    rcirs = pnb.T @ np.asarray(reci)
    vcirs = pnb.T @ np.asarray(veci)
    acirs = pnb.T @ np.asarray(aeci)

    return rcirs, vcirs, acirs


def cirs2eci(
    rcirs: ArrayLike,
    vcirs: ArrayLike,
    acirs: ArrayLike,
    ttt: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
):
    """Transforms a vector from the Celestial Intermediate Reference System (CIRS) frame
    to the ECI mean equator, mean equinox (J2000) frame using the XYS approach.

    References:
        Vallado: 2022, p. 214

    Args:
        rcirs (array_like): CIRS position vector in km
        vcirs (array_like): CIRS velocity vector in km/s
        acirs (array_like): CIRS acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Compute transformation matrix using XYS approach
    *_, pnb = iau.iau06xys(ttt, iau06arr, ddx, ddy, iau06xysarr, use_full_series)

    # Transform vectors
    reci = pnb @ np.asarray(rcirs)
    veci = pnb @ np.asarray(vcirs)
    aeci = pnb @ np.asarray(acirs)

    return reci, veci, aeci


########################################################################################
# ECI <-> TIRS Frame Conversions
########################################################################################


def eci2tirs(
    reci: ArrayLike,
    veci: ArrayLike,
    aeci: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the ECI mean equator, mean equinox (J2000) frame to the
    Terrestrial Intermediate Reference System (TIRS) frame using the XYS approach.

    References:
        Vallado: 2022, p. 213

    Args:
        reci (array_like): ECI position vector in km
        veci (array_like): ECI velocity vector in km/s
        aeci (array_like): ECI acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (rtirs, vtirs, atirs)
            rtirs (np.ndarray): TIRS position vector in km
            vtirs (np.ndarray): TIRS velocity vector in km/s
            atirs (np.ndarray): TIRS acceleration vector in km/s²
    """
    # Compute transformation matrices using XYS approach
    prec = np.eye(3)
    nut, st, _, omegaearth = compute_iau06_matrices(
        ttt, jdut1, lod, 0, 0, iau06arr, iau06xysarr, ddx, ddy, use_full_series
    )

    # Position transformation
    rtirs = st.T @ nut.T @ prec.T @ reci

    # Velocity transformation
    vtirs = st.T @ nut.T @ prec.T @ veci - np.cross(omegaearth, rtirs)

    # Acceleration transformation
    atirs = (
        st.T @ nut.T @ prec.T @ aeci
        - np.cross(omegaearth, np.cross(omegaearth, rtirs))
        - 2 * np.cross(omegaearth, vtirs)
    )

    return rtirs, vtirs, atirs


def tirs2eci(
    rtirs: ArrayLike,
    vtirs: ArrayLike,
    atirs: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    iau06arr: IAU06Array,
    iau06xysarr: IAU06xysArray,
    ddx: float = 0.0,
    ddy: float = 0.0,
    use_full_series: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Terrestrial Intermediate Reference System (TIRS)
    frame to the ECI mean equator, mean equinox (J2000) frame using the XYS approach.

    References:
        Vallado: 2022, p. 213

    Args:
        rtirs (array_like): TIRS position vector in km
        vtirs (array_like): TIRS velocity vector in km/s
        atirs (array_like): TIRS acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        iau06arr (IAU06Array): IAU 2006 data
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        ddx (float, optional): EOP correction for x in radians (default 0)
        ddy (float, optional): EOP correction for y in radians (default 0)
        use_full_series (bool, optional): Use full series for IAU 2006 XYS data

    Returns:
        tuple: (reci, veci, aeci)
            reci (np.ndarray): ECI position vector in km
            veci (np.ndarray): ECI velocity vector in km/s
            aeci (np.ndarray): ECI acceleration vector in km/s²
    """
    # Compute transformation matrices using XYS approach
    prec = np.eye(3)
    nut, st, _, omegaearth = compute_iau06_matrices(
        ttt, jdut1, lod, 0, 0, iau06arr, iau06xysarr, ddx, ddy, use_full_series
    )

    # Position transformation
    reci = prec @ nut @ st @ np.asarray(rtirs)

    # Velocity transformation
    veci = (
        prec @ nut @ st @ (np.asarray(vtirs) + np.cross(omegaearth, np.asarray(rtirs)))
    )

    # Acceleration transformation
    aeci = (
        prec
        @ nut
        @ st
        @ (
            np.asarray(atirs)
            + np.cross(omegaearth, np.cross(omegaearth, np.asarray(rtirs)))
            + 2 * np.cross(omegaearth, np.asarray(vtirs))
        )
    )

    return reci, veci, aeci


########################################################################################
# ECEF <-> PEF Frame Conversions
########################################################################################


def ecef2pef(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    xp: float,
    yp: float,
    ttt: float,
    use_iau80: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the pseudo Earth-fixed
    (PEF) frame.

    References:
        Vallado: 2022, p. 224

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ttt (float): Julian centuries of TT
        use_iau80 (bool, optional): Use IAU 1980 data for polar motion (default True)

    Returns:
        tuple: (rpef, vpef, apef)
            rpef (np.ndarray): PEF position vector in km
            vpef (np.ndarray): PEF velocity vector in km/s
            apef (np.ndarray): PEF acceleration vector in km/s²
    """
    # Compute polar motion matrix
    pm = polarm(xp, yp, ttt, use_iau80)

    # Transform vectors
    rpef = pm @ np.asarray(recef)
    vpef = pm @ np.asarray(vecef)
    apef = pm @ np.asarray(aecef)

    return rpef, vpef, apef


def pef2ecef(
    rpef: ArrayLike,
    vpef: ArrayLike,
    apef: ArrayLike,
    xp: float,
    yp: float,
    ttt: float,
    use_iau80: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the pseudo Earth-fixed (PEF) frame to the Earth-fixed
    (ECEF) frame.

    References:
        Vallado: 2022, p. 224

    Args:
        rpef (array_like): PEF position vector in km
        vpef (array_like): PEF velocity vector in km/s
        apef (array_like): PEF acceleration vector in km/s²
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ttt (float): Julian centuries of TT
        use_iau80 (bool, optional): Use IAU 1980 data for polar motion (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Compute polar motion matrix
    pm = polarm(xp, yp, ttt, use_iau80)

    # Transform vectors
    recef = pm.T @ np.asarray(rpef)
    vecef = pm.T @ np.asarray(vpef)
    aecef = pm.T @ np.asarray(apef)

    return recef, vecef, aecef


########################################################################################
# ECEF <-> TOD Frame Conversions
########################################################################################


def ecef2tod(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the true-of-date (TOD)
    frame.

    References:
        Vallado: 2022, p. 225

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (rtod, vtod, atod)
            rtod (np.ndarray): TOD position vector in km
            vtod (np.ndarray): TOD velocity vector in km/s
            atod (np.ndarray): TOD acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    _, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform position
    rpef = pm @ recef
    rtod = st @ rpef

    # Transform velocity
    vpef = pm @ vecef
    vtod = st @ (vpef + np.cross(omegaearth, rpef))

    # Transform acceleration
    atod = st @ (
        pm @ aecef
        + np.cross(omegaearth, np.cross(omegaearth, rpef))
        + 2 * np.cross(omegaearth, vpef)
    )

    return rtod, vtod, atod


def tod2ecef(
    rtod: ArrayLike,
    vtod: ArrayLike,
    atod: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the true-of-date (TOD) frame to the Earth-fixed (ECEF)
    frame.

    References:
        Vallado: 2022, p. 225

    Args:
        rtod (array_like): TOD position vector in km
        vtod (array_like): TOD velocity vector in km/s
        atod (array_like): TOD acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    _, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform position
    rpef = st.T @ rtod
    recef = pm.T @ rpef

    # Transform velocity
    vpef = st.T @ vtod - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Transform acceleration
    aecef = (
        pm.T @ (st.T @ atod)
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


########################################################################################
# ECEF <-> MOD Frame Conversions
########################################################################################


def ecef2mod(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the mean-of-date (MOD)
    frame.

    References:
        Vallado: 2022, p. 227

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (rmod, vmod, amod)
            rmod (np.ndarray): MOD position vector in km
            vmod (np.ndarray): MOD velocity vector in km/s
            amod (np.ndarray): MOD acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    _, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform position
    rpef = pm @ recef
    rmod = nut @ st @ rpef

    # Transform velocity
    vpef = pm @ vecef
    vmod = nut @ st @ (vpef + np.cross(omegaearth, rpef))

    # Transform acceleration
    temp = np.cross(omegaearth, rpef)
    amod = (
        nut
        @ st
        @ (pm @ aecef + np.cross(omegaearth, temp) + 2 * np.cross(omegaearth, vpef))
    )

    return rmod, vmod, amod


def mod2ecef(
    rmod: ArrayLike,
    vmod: ArrayLike,
    amod: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    ddpsi: float,
    ddeps: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the mean-of-date (MOD) frame to the Earth-fixed (ECEF)
    frame.

    References:
        Vallado: 2022, p. 227

    Args:
        rmod (array_like): MOD position vector in km
        vmod (array_like): MOD velocity vector in km/s
        amod (array_like): MOD acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta epsilon correction to GCRF in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Find matrices that account for various orbit effects
    _, nut, st, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, ddpsi, ddeps, iau80arr, eqeterms=eqeterms
    )

    # Transform position
    rpef = st.T @ nut.T @ rmod
    recef = pm.T @ rpef

    # Transform velocity
    vpef = st.T @ nut.T @ vmod - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Transform acceleration
    aecef = (
        pm.T @ (st.T @ nut.T @ amod)
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


########################################################################################
# ECEF <-> TEME Frame Conversions
########################################################################################


def get_teme_transform_matrices(
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau80arr: IAU80Array,
    eqeterms: bool,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute the matrices and Earth rotation vector for TEME transformations.

    Args:
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool): Add terms for ast calculation

    Returns:
        tuple: (st, pm, omegaearth)
            st (np.ndarray): Sidereal time matrix (PEF to TEME)
            pm (np.ndarray): Polar motion matrix (ECEF to PEF)
            omegaearth (np.ndarray): Earth's angular velocity vector
    """
    # Compute Greenwich Mean Sidereal Time (GMST)
    gmst = gstime(jdut1)

    # Compute omega from nutation theory
    omega = (
        125.04452222
        + (-6962890.539 * ttt + 7.455 * ttt**2 + 0.008 * ttt**3) / const.DEG2ARCSEC
    )
    omega = np.remainder(np.radians(omega), const.TWOPI)

    # Adjust GMST for geometric terms (kinematic after 1997)
    if jdut1 > 2450449.5 and eqeterms:
        gmstg = (
            gmst
            + 0.00264 * const.ARCSEC2RAD * np.sin(omega)
            + 0.000063 * const.ARCSEC2RAD * np.sin(2 * omega)
        )
    else:
        gmstg = gmst

    gmstg = np.remainder(gmstg, const.TWOPI)

    # Sidereal time matrix
    st = np.array(
        [
            [np.cos(gmstg), -np.sin(gmstg), 0],
            [np.sin(gmstg), np.cos(gmstg), 0],
            [0, 0, 1],
        ]
    )

    # Polar motion matrix and Earth's rotation vector
    *_, pm, omegaearth = calc_orbit_effects(
        ttt, jdut1, lod, xp, yp, 0, 0, iau80arr, eqeterms=eqeterms
    )

    return st, pm, omegaearth


def ecef2teme(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the true equator, mean
    mean equinox (TEME) frame.

    Results take into account the effects of sidereal time and polar motion.

    References:
        Vallado: 2022, p. 232-234

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (rteme, vteme, ateme)
            rteme (np.ndarray): TEME position vector in km
            vteme (np.ndarray): TEME velocity vector in km/s
            ateme (np.ndarray): TEME acceleration vector in km/s²
    """
    # Get common matrices and Earth rotation vector
    st, pm, omegaearth = get_teme_transform_matrices(
        ttt, jdut1, lod, xp, yp, iau80arr, eqeterms
    )

    # Transform position
    rpef = pm @ recef
    rteme = st @ rpef

    # Transform velocity
    vpef = pm @ vecef
    vteme = st @ (vpef + np.cross(omegaearth, rpef))

    # Transform acceleration
    ateme = st @ (
        pm @ aecef
        + np.cross(omegaearth, np.cross(omegaearth, rpef))
        + 2 * np.cross(omegaearth, vpef)
    )

    return rteme, vteme, ateme


def teme2ecef(
    rteme: ArrayLike,
    vteme: ArrayLike,
    ateme: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
    iau80arr: IAU80Array,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the true equator, mean equinox (TEME) frame to the
    Earth-fixed (ECEF) frame.

    Results take into account the effects of sidereal time and polar motion.

    References:
        Vallado: 2022, p. 232-234

    Args:
        rteme (array_like): TEME position vector in km
        vteme (array_like): TEME velocity vector in km/s
        ateme (array_like): TEME acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        iau80arr (IAU80Array): IAU 1980 data for nutation
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Get common matrices and Earth rotation vector
    st, pm, omegaearth = get_teme_transform_matrices(
        ttt, jdut1, lod, xp, yp, iau80arr, eqeterms
    )

    # Transform position
    rpef = st.T @ rteme
    recef = pm.T @ rpef

    # Transform velocity
    vpef = st.T @ vteme - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Transform acceleration
    aecef = pm.T @ (
        st.T @ ateme
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


########################################################################################
# ECEF <-> CIRS Frame Conversions
########################################################################################


def ecef2cirs(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the Celestial
    Intermediate Reference System (CIRS) frame.

    References:
        Vallado: 2022, p. 214

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians

    Returns:
        tuple: (rcirs, vcirs, acirs)
            rcirs (np.ndarray): CIRS position vector in km
            vcirs (np.ndarray): CIRS velocity vector in km/s
            acirs (np.ndarray): CIRS acceleration vector in km/s²
    """
    # Compute transformation matrices
    st, _ = sidereal(jdut1, 0, 0, 0, lod, use_iau80=False)
    omegaearth = calc_omegaearth(lod)
    pm = polarm(xp, yp, ttt, use_iau80=False)

    # Transform position
    rpef = pm @ np.asarray(recef)
    rcirs = st @ rpef

    # Transform velocity
    vpef = pm @ np.asarray(vecef)
    vcirs = st @ (vpef + np.cross(omegaearth, rpef))

    # Transform acceleration
    acirs = st @ (
        pm @ np.asarray(aecef)
        + np.cross(omegaearth, np.cross(omegaearth, rpef))
        + 2 * np.cross(omegaearth, vpef)
    )

    return rcirs, vcirs, acirs


def cirs2ecef(
    rcirs: ArrayLike,
    vcirs: ArrayLike,
    acirs: ArrayLike,
    ttt: float,
    jdut1: float,
    lod: float,
    xp: float,
    yp: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Celestial Intermediate Reference System (CIRS) frame
    to the Earth-fixed (ECEF) frame.

    References:
        Vallado: 2022, p. 214

    Args:
        rcirs (array_like): CIRS position vector in km
        vcirs (array_like): CIRS velocity vector in km/s
        acirs (array_like): CIRS acceleration vector in km/s²
        ttt (float): Julian centuries of TT
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lod (float): Excess length of day in seconds
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    # Compute transformation matrices
    st, _ = sidereal(jdut1, 0, 0, 0, lod, use_iau80=False)
    omegaearth = calc_omegaearth(lod)
    pm = polarm(xp, yp, ttt, use_iau80=False)

    # Transform position
    rpef = st.T @ rcirs
    recef = pm.T @ rpef

    # Transform velocity
    vpef = st.T @ vcirs - np.cross(omegaearth, rpef)
    vecef = pm.T @ vpef

    # Transform acceleration
    aecef = (
        pm.T @ (st.T @ acirs)
        - np.cross(omegaearth, np.cross(omegaearth, rpef))
        - 2 * np.cross(omegaearth, vpef)
    )

    return recef, vecef, aecef


########################################################################################
# ECEF <-> TIRS Frame Conversions
########################################################################################


def ecef2tirs(
    recef: ArrayLike,
    vecef: ArrayLike,
    aecef: ArrayLike,
    xp: float,
    yp: float,
    ttt: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Earth-fixed (ECEF) frame to the Terrestrial
    Intermediate Reference System (TIRS) frame.

    References:
        Vallado: 2022, p. 213

    Args:
        recef (array_like): ECEF position vector in km
        vecef (array_like): ECEF velocity vector in km/s
        aecef (array_like): ECEF acceleration vector in km/s²
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ttt (float): Julian centuries of TT

    Returns:
        tuple: (rtirs, vtirs, atirs)
            rtirs (np.ndarray): TIRS position vector in km
            vtirs (np.ndarray): TIRS velocity vector in km/s
            atirs (np.ndarray): TIRS acceleration vector in km/s²
    """
    return ecef2pef(recef, vecef, aecef, xp, yp, ttt, use_iau80=False)


def tirs2ecef(
    rtirs: ArrayLike,
    vtirs: ArrayLike,
    atirs: ArrayLike,
    xp: float,
    yp: float,
    ttt: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Transforms a vector from the Terrestrial Intermediate Reference System (TIRS)
    frame to the Earth-fixed (ECEF) frame.

    References:
        Vallado: 2022, p. 213

    Args:
        rtirs (array_like): TIRS position vector in km
        vtirs (array_like): TIRS velocity vector in km/s
        atirs (array_like): TIRS acceleration vector in km/s²
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ttt (float): Julian centuries of TT

    Returns:
        tuple: (recef, vecef, aecef)
            recef (np.ndarray): ECEF position vector in km
            vecef (np.ndarray): ECEF velocity vector in km/s
            aecef (np.ndarray): ECEF acceleration vector in km/s²
    """
    return pef2ecef(rtirs, vtirs, atirs, xp, yp, ttt, use_iau80=False)



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 25 June 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
import math
from dataclasses import dataclass
from typing import Tuple

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.optimize import root_scalar

from .data import IAU80Array
from ...constants import ARCSEC2RAD, DEG2ARCSEC, TWOPI, HR2SEC
from ...mathtime.vector import rot1mat, rot2mat


# Initialize logger
logger = logging.getLogger(__name__)


@dataclass
class FundArgs:
    # fmt: off
    """Data class for Delaunay fundamental arguments."""
    l: float = 0.0         # delaunay element in radians
    l1: float = 0.0        # delaunay element in radians
    f: float = 0.0         # delaunay element in radians
    d: float = 0.0         # delaunay element in radians
    omega: float = 0.0     # delaunay element in radians
    lonmer: float = 0.0    # longitude of Mercury in radians
    lonven: float = 0.0    # longitude of Venus in radians
    lonear: float = 0.0    # longitude of Earth in radians
    lonmar: float = 0.0    # longitude of Mars in radians
    lonjup: float = 0.0    # longitude of Jupiter in radians
    lonsat: float = 0.0    # longitude of Saturn in radians
    lonurn: float = 0.0    # longitude of Uranus in radians
    lonnep: float = 0.0    # longitude of Neptune in radians
    precrate: float = 0.0  # precession rate in radians per Julian century


def fundarg(ttt: float, opt: str) -> FundArgs:
    """Calculates the Delaunay variables and planetary values for several theories.

    References:
        Vallado: 2022, p. 210-212, 226

    Args:
        ttt (float): Julian centuries of TT
        opt (str): Method option ('06', '02', '96', or '80')

    Returns:
        FundArgs: Delaunay fundamental arguments and planetary values

    TODO:
        - Implement commented out methods (from m-file)?
        - Use enums instead of strings for option/method
    """

    def calc_delunay_elem(ttt, coeffs):
        """Delaunay fundamental arguments formed in arcsec, converted to deg"""
        return (
            (((coeffs[0] * ttt + coeffs[1]) * ttt + coeffs[2]) * ttt + coeffs[3]) * ttt
            + coeffs[4]
        ) / DEG2ARCSEC

    def calc_delunay_elem_80(ttt, coeffs, extra):
        return (
            ((coeffs[0] * ttt + coeffs[1]) * ttt + coeffs[2]) * ttt
        ) / DEG2ARCSEC + extra

    # Initialize data object
    fundargs = FundArgs()

    # Determine coefficients from IAU 2006 nutation theory
    if opt == "06":
        # Delaunay fundamental arguments in deg
        fundargs.l = calc_delunay_elem(  # noqa
            ttt, [-0.0002447, 0.051635, 31.8792, 1717915923.2178, 485868.249036]
        )
        fundargs.l1 = calc_delunay_elem(
            ttt, [-0.00001149, 0.000136, -0.5532, 129596581.0481, 1287104.793048]
        )
        fundargs.f = calc_delunay_elem(
            ttt, [0.00000417, -0.001037, -12.7512, 1739527262.8478, 335779.526232]
        )
        fundargs.d = calc_delunay_elem(
            ttt, [-0.00003169, 0.006593, -6.3706, 1602961601.209, 1072260.703692]
        )
        fundargs.omega = calc_delunay_elem(
            ttt, [-0.00005939, 0.007702, 7.4722, -6962890.5431, 450160.398036]
        )

        # Planetary arguments in deg (from TN-36)
        fundargs.lonmer = np.mod((4.402608842 + 2608.7903141574 * ttt), TWOPI)
        fundargs.lonven = np.mod((3.176146697 + 1021.3285546211 * ttt), TWOPI)
        fundargs.lonear = np.mod((1.753470314 + 628.3075849991 * ttt), TWOPI)
        fundargs.lonmar = np.mod((6.203480913 + 334.06124267 * ttt), TWOPI)
        fundargs.lonjup = np.mod((0.599546497 + 52.9690962641 * ttt), TWOPI)
        fundargs.lonsat = np.mod((0.874016757 + 21.329910496 * ttt), TWOPI)
        fundargs.lonurn = np.mod((5.481293872 + 7.4781598567 * ttt), TWOPI)
        fundargs.lonnep = np.mod((5.311886287 + 3.8133035638 * ttt), TWOPI)
        fundargs.precrate = (0.02438175 + 0.00000538691 * ttt) * ttt

    # Determine coefficients from IAU 2000b theory
    elif opt == "02":
        # Delaunay fundamental arguments in deg
        # Planetary longitues and precession rates remain at zero
        fundargs.l = 134.96340251 + (1717915923.2178 * ttt) / DEG2ARCSEC  # noqa
        fundargs.l1 = 357.52910918 + (129596581.0481 * ttt) / DEG2ARCSEC
        fundargs.f = 93.27209062 + (1739527262.8478 * ttt) / DEG2ARCSEC
        fundargs.d = 297.85019547 + (1602961601.209 * ttt) / DEG2ARCSEC
        fundargs.omega = 125.04455501 + (-6962890.5431 * ttt) / DEG2ARCSEC

    # Determine coefficients from IAU 1996 theory
    elif opt == "96":
        # Delaunay fundamental arguments in deg
        fundargs.l = (  # noqa
            calc_delunay_elem(ttt, [-0.0002447, 0.051635, 31.8792, 1717915923.2178, 0])
            + 134.96340251
        )
        fundargs.l1 = (
            calc_delunay_elem(ttt, [-0.00001149, -0.000136, -0.5532, 129596581.0481, 0])
            + 357.52910918
        )
        fundargs.f = (
            calc_delunay_elem(ttt, [0.00000417, 0.001037, -12.7512, 1739527262.8478, 0])
            + 93.27209062
        )
        fundargs.d = (
            calc_delunay_elem(ttt, [-0.00003169, 0.006593, -6.3706, 1602961601.2090, 0])
            + 297.85019547
        )
        fundargs.omega = (
            calc_delunay_elem(ttt, [-0.00005939, 0.007702, 7.4722, -6962890.2665, 0])
            + 125.04455501
        )

        # Planetary arguments in deg
        fundargs.lonven = 181.979800853 + 58517.8156748 * ttt
        fundargs.lonear = 100.466448494 + 35999.3728521 * ttt
        fundargs.lonmar = 355.433274605 + 19140.299314 * ttt
        fundargs.lonjup = 34.3514839 + 3034.90567464 * ttt
        fundargs.lonsat = 50.0774713998 + 1222.11379404 * ttt
        fundargs.precrate = 1.39697137214 * ttt + 0.0003086 * ttt**2

    # Determine coefficients from IAU 1980 theory
    elif opt == "80":
        # Delaunay fundamental arguments in deg
        fundargs.l = calc_delunay_elem_80(  # noqa
            ttt, [0.064, 31.31, 1717915922.633], 134.96298139
        )
        fundargs.l1 = calc_delunay_elem_80(
            ttt, [-0.012, -0.577, 129596581.224], 357.52772333
        )
        fundargs.f = calc_delunay_elem_80(
            ttt, [0.011, -13.257, 1739527263.137], 93.27191028
        )
        fundargs.d = calc_delunay_elem_80(
            ttt, [0.019, -6.891, 1602961601.328], 297.85036306
        )
        fundargs.omega = calc_delunay_elem_80(
            ttt, [0.008, 7.455, -6962890.539], 125.04452222
        )

        # Planetary arguments in deg
        fundargs.lonmer = 252.3 + 149472 * ttt
        fundargs.lonven = 179.9 + 58517.8 * ttt
        fundargs.lonear = 98.4 + 35999.4 * ttt
        fundargs.lonmar = 353.3 + 19140.3 * ttt
        fundargs.lonjup = 32.3 + 3034.9 * ttt
        fundargs.lonsat = 48 + 1222.1 * ttt
    else:
        raise ValueError(
            "Method must be one of the following: '06', '02', '96', or '80'"
        )

    # Convert values to radians
    twopi_deg = np.degrees(TWOPI)
    fundargs.l = float(np.radians(np.mod(fundargs.l, twopi_deg)))  # noqa
    fundargs.l1 = float(np.radians(np.mod(fundargs.l1, twopi_deg)))
    fundargs.f = float(np.radians(np.mod(fundargs.f, twopi_deg)))
    fundargs.d = float(np.radians(np.mod(fundargs.d, twopi_deg)))
    fundargs.omega = float(np.radians(np.mod(fundargs.omega, twopi_deg)))
    if not opt == "06":
        fundargs.lonmer = float(np.radians(np.mod(fundargs.lonmer, twopi_deg)))
        fundargs.lonven = float(np.radians(np.mod(fundargs.lonven, twopi_deg)))
        fundargs.lonear = float(np.radians(np.mod(fundargs.lonear, twopi_deg)))
        fundargs.lonmar = float(np.radians(np.mod(fundargs.lonmar, twopi_deg)))
        fundargs.lonjup = float(np.radians(np.mod(fundargs.lonjup, twopi_deg)))
        fundargs.lonsat = float(np.radians(np.mod(fundargs.lonsat, twopi_deg)))
        fundargs.lonurn = float(np.radians(np.mod(fundargs.lonurn, twopi_deg)))
        fundargs.lonnep = float(np.radians(np.mod(fundargs.lonnep, twopi_deg)))
        fundargs.precrate = float(np.radians(np.mod(fundargs.precrate, twopi_deg)))

    return fundargs


def precess(ttt: float, opt: str) -> Tuple[np.ndarray, float, float, float, float]:
    """Calculates the transformation matrix that accounts for the effects of
    precession. Both the 1980 and 2006 IAU theories are handled, as well as the
    FK B1950 theory.

    References:
        Vallado: 2022, p. 219, 227-229

    Args:
        ttt (float): Julian centuries of Terrestrail Time (TT)
        opt (str): Method option ('50', '80', or '06')
                   '50' = FK4 B1950
                   '80' = IAU 1980
                   '06' = IAU 2006

    Returns:
        tuple: (prec, psia, wa, ea, xa)
            prec (np.array): Transformation matrix for MOD to J2000
            psia (float): Canonical precession angle in radians
            wa (float): Canonical precession angle in radians
            ea (float): Canonical precession angle in radians
            xa (float): Canonical precession angle in radians

    TODO:
        - Implement commented out methods (from m-file)?
        - Use enums instead of strings for option/method
    """

    def calc_prec_angle(ttt, coeffs):
        return (
            (((coeffs[0] * ttt + coeffs[1]) * ttt + coeffs[2]) * ttt + coeffs[3]) * ttt
            + coeffs[4]
        ) * ttt

    # Initialize some variables
    ttt2, ttt3 = ttt**2, ttt**3
    prec = np.eye(3)

    # FK4 B1950 precession angles
    if opt == "50":
        # Commenting these out because they seem important but not used
        # TODO: Decide if these need to be used instead of definitions below
        # psia = 50.3708 + 0.0050 * ttt
        # wa = 0.0
        # ea = 84428.26 - 46.845 * ttt - 0.00059 * ttt2 + 0.00181 * ttt3
        xa = 0.1247 - 0.0188 * ttt

        # GTDS pg 3-17 using days from 1950 - avoids long precession constants
        zeta = 2304.9969 * ttt + 0.302 * ttt2 + 0.01808 * ttt3
        theta = 2004.298 * ttt - 0.425936 * ttt2 - 0.0416 * ttt3
        z = 2304.9969 * ttt + 1.092999 * ttt2 + 0.0192 * ttt3

        # ttt is tropical centuries from 1950 (36524.22 days)
        prec[0, 0] = 1 - 2.9696e-4 * ttt2 - 1.3e-7 * ttt3
        prec[0, 1] = 2.234941e-2 * ttt + 6.76e-6 * ttt2 - 2.21e-6 * ttt3
        prec[0, 2] = 9.7169e-3 * ttt - 2.07e-6 * ttt2 - 9.6e-7 * ttt3
        prec[1, 0] = -prec[0, 1]
        prec[1, 1] = 1 - 2.4975e-4 * ttt2 - 1.5e-7 * ttt3
        prec[1, 2] = -1.0858e-4 * ttt2
        prec[2, 0] = -prec[0, 2]
        prec[2, 1] = prec[1, 2]
        prec[2, 2] = 1 - 4.721e-5 * ttt2

        # Pass these back out for testing
        # TODO: decide if these need to be removed
        psia, wa, ea = zeta, theta, z

    # IAU 80 precession angles
    elif opt == "80":
        psia = 5038.7784 * ttt - 1.07259 * ttt2 - 0.001147 * ttt3
        wa = 84381.448 + 0.05127 * ttt2 - 0.007726 * ttt3
        ea = 84381.448 - 46.8150 * ttt - 0.00059 * ttt2 + 0.001813 * ttt3
        xa = 10.5526 * ttt - 2.38064 * ttt2 - 0.001125 * ttt3

        zeta = 2306.2181 * ttt + 0.30188 * ttt2 + 0.017998 * ttt3
        theta = 2004.3109 * ttt - 0.42665 * ttt2 - 0.041833 * ttt3
        z = 2306.2181 * ttt + 1.09468 * ttt2 + 0.018203 * ttt3

    # IAU 06 precession angles
    elif opt == "06":
        oblo = 84381.406
        psia = calc_prec_angle(
            ttt, [-0.0000000951, 0.000132851, -0.00114045, -1.0790069, 5038.481507]
        )
        wa = (
            calc_prec_angle(
                ttt, [0.0000003337, -0.000000467, -0.00772503, 0.0512623, -0.025754]
            )
            + oblo
        )
        ea = (
            calc_prec_angle(
                ttt, [-0.0000000434, -0.000000576, 0.0020034, -0.0001831, -46.836769]
            )
            + oblo
        )
        xa = calc_prec_angle(
            ttt, [-0.000000056, 0.000170663, -0.00121197, -2.3814292, 10.556403]
        )
        zeta = (
            calc_prec_angle(
                ttt, [-0.0000003173, -0.000005971, 0.01801828, 0.2988499, 2306.083227]
            )
            + 2.650545
        )
        theta = calc_prec_angle(
            ttt, [-0.0000001274, -0.000007089, -0.04182264, -0.4294934, 2004.191903]
        )
        z = (
            calc_prec_angle(
                ttt, [0.0000002904, -0.000028596, 0.01826837, 1.0927348, 2306.077181]
            )
            - 2.650545
        )
    else:
        raise ValueError("Method must be one of the following: '50', '80', or '06'")

    # Convert units to radians
    zeta *= ARCSEC2RAD
    theta *= ARCSEC2RAD
    z *= ARCSEC2RAD

    # IAU precession angles
    if opt in ["80", "06"]:
        coszeta = np.cos(zeta)
        sinzeta = np.sin(zeta)
        costheta = np.cos(theta)
        sintheta = np.sin(theta)
        cosz = np.cos(z)
        sinz = np.sin(z)

        # Form matrix MOD to J2000
        prec[0, 0] = coszeta * costheta * cosz - sinzeta * sinz
        prec[0, 1] = coszeta * costheta * sinz + sinzeta * cosz
        prec[0, 2] = coszeta * sintheta
        prec[1, 0] = -sinzeta * costheta * cosz - coszeta * sinz
        prec[1, 1] = -sinzeta * costheta * sinz + coszeta * cosz
        prec[1, 2] = -sinzeta * sintheta
        prec[2, 0] = -sintheta * cosz
        prec[2, 1] = -sintheta * sinz
        prec[2, 2] = costheta

    return prec, psia * ARCSEC2RAD, wa * ARCSEC2RAD, ea * ARCSEC2RAD, xa * ARCSEC2RAD


def _get_nutation_parameters(ttt, iau80arr, model):
    # Fundamental arguments
    fundargs = fundarg(ttt, model)

    # Calculate nutation parameters
    deltapsi, deltaeps = 0, 0
    for i in range(len(iau80arr.iar80)):
        tempval = (
            iau80arr.iar80[i, 0] * fundargs.l
            + iau80arr.iar80[i, 1] * fundargs.l1
            + iau80arr.iar80[i, 2] * fundargs.f
            + iau80arr.iar80[i, 3] * fundargs.d
            + iau80arr.iar80[i, 4] * fundargs.omega
        )
        deltapsi += (iau80arr.rar80[i, 0] + iau80arr.rar80[i, 1] * ttt) * np.sin(
            tempval
        )
        deltaeps += (iau80arr.rar80[i, 2] + iau80arr.rar80[i, 3] * ttt) * np.cos(
            tempval
        )

    return deltapsi, deltaeps, fundargs.omega


def _build_nutation_matrix(deltapsi, meaneps, trueeps):
    cospsi, sinpsi = np.cos(deltapsi), np.sin(deltapsi)
    coseps, sineps = np.cos(meaneps), np.sin(meaneps)
    costrueeps, sintrueeps = np.cos(trueeps), np.sin(trueeps)

    nut = np.zeros((3, 3))
    nut[0, 0] = cospsi
    nut[0, 1] = costrueeps * sinpsi
    nut[0, 2] = sintrueeps * sinpsi
    nut[1, 0] = -coseps * sinpsi
    nut[1, 1] = costrueeps * coseps * cospsi + sintrueeps * sineps
    nut[1, 2] = sintrueeps * coseps * cospsi - sineps * costrueeps
    nut[2, 0] = -sineps * sinpsi
    nut[2, 1] = costrueeps * sineps * cospsi - sintrueeps * coseps
    nut[2, 2] = sintrueeps * sineps * cospsi + costrueeps * coseps

    return nut


def nutation(
    ttt: float, ddpsi: float, ddeps: float, iau80arr: IAU80Array
) -> Tuple[float, float, float, float, np.ndarray]:
    """Calculates the transformation matrix that accounts for the effects of nutation.

    References:
        Vallado: 2022, p. 225-227

    Args:
        ttt (float): Julian centuries of TT
        ddpsi (float): Delta psi correction to GCRF in radians
        ddeps (float): Delta eps correction to GCRF in radians
        iau80arr (IAU80Array): Data object containing the nutation matrices

    Returns:
        tuple: (deltapsi, trueeps, meaneps, omega, nut)
            deltapsi (float): Nutation angle in radians
            trueeps (float): True obliquity of the ecliptic in radians
            meaneps (float): Mean obliquity of the ecliptic in radians
            omega (float): Delaunay element in radians
            nut (np.ndarray): Transformation matrix for TOD - MOD
    """
    # Mean obliquity of the ecliptic
    meaneps = -46.815 * ttt - 0.00059 * ttt**2 + 0.001813 * ttt**3 + 84381.448
    meaneps = float(np.radians(np.remainder(meaneps / DEG2ARCSEC, np.degrees(TWOPI))))

    # Calculate nutation parameters
    deltapsi, deltaeps, omega = _get_nutation_parameters(ttt, iau80arr, "80")

    # Add corrections
    deltapsi = math.remainder(deltapsi + ddpsi, TWOPI)
    deltaeps = math.remainder(deltaeps + ddeps, TWOPI)
    trueeps = meaneps + deltaeps

    # Construct nutation rotation matrix
    nut = _build_nutation_matrix(deltapsi, meaneps, trueeps)

    return deltapsi, trueeps, meaneps, omega, nut


def nutation_qmod(
    ttt: float, iau80arr: IAU80Array, use_eutelsat_approx: bool = False
) -> Tuple[float, float, float, float, np.ndarray]:
    """Calculates the transformation matrix that accounts for the effects of nutation
    within the Quasi Mean-of-Date (QMOD) paradigm.

    References:
        Vallado: 2022, p. 225-227

    Args:
        ttt (float): Julian centuries of TT
        iau80arr (IAU80Array): Data object containing the nutation matrices
        use_eutelsat_approx (bool, optional): Whether to use the Eutelsat
                                              approximation (defaults to False)

    Returns:
        tuple: (deltapsi, trueeps, meaneps, omega, nut)
            deltapsi (float): Nutation angle in radians
            trueeps (float): True obliquity of the ecliptic in radians
            meaneps (float): Mean obliquity of the ecliptic in radians
            omega (float): Delaunay element in radians
            nut (np.ndarray): Transformation matrix for TOD - MOD
    """
    # Mean obliquity of the ecliptic
    meaneps = np.radians(84381.448 / HR2SEC) % TWOPI

    # Calculate nutation parameters
    deltapsi, deltaeps, omega = _get_nutation_parameters(ttt, iau80arr, "96")

    # Add corrections
    deltapsi = math.remainder(deltapsi, TWOPI)
    deltaeps = math.remainder(deltaeps, TWOPI)
    trueeps = meaneps + deltaeps

    # Construct nutation rotation matrix
    if use_eutelsat_approx:
        # Eutelsat approximation
        n1 = rot1mat(deltaeps)
        n2 = rot2mat(-deltapsi * np.sin(meaneps))
        nut = n2 @ n1
    else:
        nut = _build_nutation_matrix(deltapsi, meaneps, trueeps)

    return deltapsi, trueeps, meaneps, omega, nut


def polarm(xp: float, yp: float, ttt: float, use_iau80: bool = True) -> np.ndarray:
    """Calculate the transformation matrix that accounts for polar motion.

    References:
        Vallado: 2022, p. 213, 224

    Both the 1980 and 2000 theories are handled. Note that the rotation order
    is different between 1980 and 2000.

    Args:
        xp (float): Polar motion coefficient in radians
        yp (float): Polar motion coefficient in radians
        ttt (float): Julian centuries of TT (only used in IAU 2000 method)
        use_iau80 (bool, optional): Whether to use the IAU 1980 method instead
                                    of IAU 2000 method (defaults to True)

    Returns:
        pm (np.ndarray): Transformation matrix for ECEF to PEF
    """
    cosxp, sinxp = np.cos(xp), np.sin(xp)
    cosyp, sinyp = np.cos(yp), np.sin(yp)

    # Use IAU 1980 theory
    if use_iau80:
        pm = np.array(
            [
                [cosxp, 0, -sinxp],
                [sinxp * sinyp, cosyp, cosxp * sinyp],
                [sinxp * cosyp, -sinyp, cosxp * cosyp],
            ]
        )
    # Use IAU 2000 theory
    else:
        # −47e-6 corresponds to a constant drift in the Terrestrial
        # Intermediate Origin (TIO) locator 𝑠′, which is approximately −47
        # microarcseconds per century (applied to the IAU 2000 theory)
        # See: https://pyerfa.readthedocs.io/en/latest/api/erfa.pom00.html
        # TODO: consider using pyerfa for this
        sp = -47e-6 * ttt * ARCSEC2RAD
        cossp, sinsp = np.cos(sp), np.sin(sp)

        pm = np.array(
            [
                [
                    cosxp * cossp,
                    -cosyp * sinsp + sinyp * sinxp * cossp,
                    -sinyp * sinsp - cosyp * sinxp * cossp,
                ],
                [
                    cosxp * sinsp,
                    cosyp * cossp + sinyp * sinxp * sinsp,
                    sinyp * cossp - cosyp * sinxp * sinsp,
                ],
                [sinxp, -sinyp * cosxp, cosyp * cosxp],
            ]
        )

    return pm


def ap_kp_table() -> Tuple[np.ndarray, np.ndarray]:
    """Generates the Kp to Ap index conversion table.

    References:
        Vallado: 2022, p. 562, Table 8-3

    Returns:
        tuple: (ap, kp)
            ap (np.ndarray): Ap index values
            kp (np.ndarray): Kp index values
    """
    # Define Kp and Ap arrays
    # fmt: off
    ap = np.array(
        [0, 2, 3, 4, 5, 6, 7, 9, 12, 15, 18, 22, 27, 32,
         39, 48, 56, 67, 80, 94, 111, 132, 154, 179, 207, 236, 300, 400]
    )

    kp = np.array(
        [0, 0.33333, 0.66667, 1, 1.33333, 1.66667, 2, 2.33333, 2.66667, 3, 3.33333,
         3.66667, 4, 4.33333, 4.66667, 5, 5.33333, 5.66667, 6, 6.33333, 6.66667, 7,
         7.33333, 7.66667, 8, 8.33333, 8.66667, 9]
    )
    # fmt: on

    return ap, kp


def kp2ap(kpin: float) -> float | None:
    """Converts Kp index to Ap index using cubic spline interpolation.

    References:
        Vallado: 2022, p. 560-562, Table 8-3

    Args:
        kpin (float): Kp index value

    Returns:
        float: Corresponding Ap index value, or None if out of bounds
    """
    # Get Ap and Kp arrays
    ap, kp = ap_kp_table()

    # Ensure kpin is within bounds
    if kpin < kp[0] or kpin > kp[-1]:
        logger.warning("Kp index out of bounds for conversion to Ap index.")
        return None

    return float(CubicSpline(kp, ap)(kpin))


def ap2kp(apin: float) -> float | None:
    """Converts Ap index to Kp index using cubic spline interpolation and root-finding.

    References:
        Vallado: 2022, p. 560-562, Table 8-3

    Args:
        apin (float): Ap index value

    Returns:
        float: Corresponding Kp index value, or None if out of bounds

    Notes:
        - Root-finding is used for Ap-to-Kp conversion due to the non-linear and uneven
          spacing in the Ap scale.
    """
    # Get Ap and Kp arrays
    ap, kp = ap_kp_table()

    # Ensure apin is within bounds
    if apin < ap[0] or apin > ap[-1]:
        logger.warning("Ap index out of bounds for conversion to Kp index.")
        return None

    # Define the spline from Kp to Ap
    spline = CubicSpline(kp, ap)

    # Root-finding to solve: spline(kp) - apin = 0
    def func(kp_val):
        return spline(kp_val) - apin

    # Root-finding between valid Kp bounds
    result = root_scalar(func, bracket=[kp[0], kp[-1]], method="brentq")

    if result.converged:
        return result.root
    else:
        logger.error("Root finding did not converge.")
        return None
